#include <Eigen/Sparse>
#include <iostream>
#include "writer.hpp"
#include <cmath>
#include <Eigen/SparseCholesky>
#include <stdexcept>

//! Sparse Matrix type. Makes using this type easier.
typedef Eigen::SparseMatrix<double> SparseMatrix;

//! Used for filling the sparse matrix.
typedef Eigen::Triplet<double> Triplet;

//! Vector type
typedef Eigen::VectorXd Vector;


//! Create the 1D Poisson matrix
//! @param[out] A will contain the Poisson matrix
//! @param[in] N the number of interior points
void createPoissonMatrix(SparseMatrix& A, int N) {
    A.resize(N, N);
    double h=1./(N+1);
    std::vector<Triplet> triplets;
    triplets.reserve(N + 2 * N - 2);
    for (int i = 0; i < N; ++i) {
        triplets.push_back(Triplet(i, i, 2./(h*h)));
        if (i > 0) {
            triplets.push_back(Triplet(i, i - 1, -1./(h*h)));
        }
        if (i < N - 1){
            triplets.push_back(Triplet(i, i + 1, -1./(h*h)));
        }
    }

    A.setFromTriplets(triplets.begin(), triplets.end());
}

/// Uses the explicit Euler method to compute u from time 0 to time T 
///
/// @param[out] u at all time steps up to time T, each column corresponding to a time step (including the initial condition as first column)
/// @param[out] time the time levels
/// @param[in] u0 the initial data, as column vector
/// @param[in] dt the time step size
/// @param[in] T the final time at which to compute the solution (which we assume to be a multiple of dt)
/// @param[in] N the number of interior grid points
///

void explicitEuler(Eigen::MatrixXd & u, Vector & time,
	const Vector u0, double dt, double T, int N) {
  const unsigned int nsteps = round(T/dt);
  u.resize(N,nsteps+1);
  time.resize(nsteps+1);
  /* Initialize A */
  SparseMatrix A;
  createPoissonMatrix(A,N);
  /* Initialize u */
  u.col(0)<<u0;
  time[0]=0.;
  for(unsigned k=0; k<nsteps; k++)
    {
 	u.col(k+1)=u.col(k)-dt*A*u.col(k);
   	time[k+1]=(k+1)*dt;
    }
}

/// Uses the Crank-Nicolson method to compute u from time 0 to time T 
///
/// @param[out] u at all time steps up to time T, each column corresponding to a time step (including the initial condition as first column)
/// @param[out] time the time levels
/// @param[in] u0 the initial data, as column vector
/// @param[in] dt the time step size
/// @param[in] T the final time at which to compute the solution (which we assume to be a multiple of dt)
/// @param[in] N the number of interior grid points
///

void CrankNicolson(Eigen::MatrixXd & u, Vector & time,
	const Vector u0, double dt, double T, int N) {
  const unsigned int nsteps = round(T/dt);
  u.resize(N,nsteps+1);
  time.resize(nsteps+1);
  /* Initialize A */
  SparseMatrix A;
  createPoissonMatrix(A,N);
  SparseMatrix B(N,N);
  B.setIdentity();
  B+=dt/2.*A;
  /* Initialize u */
  u.col(0)<<u0;
  time[0]=0.;
  /* Initialize solver and compute Cholesky decomposition of B (Note: since dt is constant, the matrix B is the same for all timesteps)*/
  Eigen::SimplicialLDLT<SparseMatrix> solver;
  solver.compute(B);
  for(unsigned k=0; k<nsteps; k++)
    {
      u.col(k+1)=solver.solve(u.col(k)-dt/2*A*u.col(k));
      time[k+1]=(k+1)*dt;
    }
}

double U0(double x) {
  return std::min(2*x,2-2*x);
}

int main(int, char**) {
    double T = 0.3;
    double dt = 0.0002; // Change this for explicit / implicit time stepping comparison
    int N=40;
    Vector u0(N);
    double h=1./(N+1);
    /* Initialize u0 */
    for(int i=0;i<u0.size();i++)
      u0[i] = U0(h*(i+1));

    Eigen::MatrixXd u_euler;
    Vector time;
    explicitEuler(u_euler,time,u0,dt,T,N);
    writeToFile("time.txt",time);
    writeMatrixToFile("u_explEuler.txt", u_euler);
    Eigen::MatrixXd u_cn;
    CrankNicolson(u_cn,time,u0,dt,T,N);
    writeMatrixToFile("u_cn.txt",u_cn);
}




